/*
 * Copyright (C) 2011-2012 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * This file originates from The Android Open Source Project
 * and contains some modifications which are (C) COPYRIGHT 2014 ARM Limited
 */

#include "../cpu_ref/rsd_cpu.h"

#include "rsdCore.h"

#include "rsdBcc.h"
#include "rsdAllocation.h"

#include "rsContext.h"
#include "rsElement.h"
#include "rsScriptC.h"

#if !defined(RS_SERVER) && !defined(RS_COMPATIBILITY_LIB)
#include "utils/Vector.h"
#include "utils/Timers.h"
#include "utils/StopWatch.h"
#endif
// ARM - include DrvScript structure definition
#include "rsdClBcc.h"

using namespace android;
using namespace android::renderscript;


bool rsdScriptInit(const Context *rsc,
                     ScriptC *script,
                     char const *resName,
                     char const *cacheDir,
                     uint8_t const *bitcode,
                     size_t bitcodeSize,
                     uint32_t flags) {
    RsdHal *dc = (RsdHal *)rsc->mHal.drv;
    RsdCpuReference::CpuScript * cs = dc->mCpuRef->createScript(script, resName, cacheDir,
                                                                bitcode, bitcodeSize, flags);
    if (cs == NULL) {
        return false;
    }
    cs->populateScript(script);
    // ARM - create storage for DrvScript structure
    DrvScript *drv = (DrvScript *)calloc(1, sizeof(DrvScript));
    if (drv == NULL) {
        delete cs;
        return false;
    }

    drv->cs = cs;
    script->mHal.drv = drv;

    return true;
}

bool rsdInitIntrinsic(const Context *rsc, Script *s, RsScriptIntrinsicID iid, Element *e) {
    RsdHal *dc = (RsdHal *)rsc->mHal.drv;
    RsdCpuReference::CpuScript * cs = dc->mCpuRef->createIntrinsic(s, iid, e);
    if (cs == NULL) {
        return false;
    }
    cs->populateScript(s);

    // ARM - create storage for DrvScript structure
    DrvScript *drv = (DrvScript *)calloc(1, sizeof(DrvScript));
    if (drv == NULL) {
        delete cs;
        return false;
    }

    drv->cs = cs;
    s->mHal.drv = drv;

    return true;
}

void rsdScriptInvokeForEach(const Context *rsc,
                            Script *s,
                            uint32_t slot,
                            const Allocation * ain,
                            Allocation * aout,
                            const void * usr,
                            size_t usrLen,
                            const RsScriptCall *sc) {

    // ARM - load CpuScript * from DrvScript structure
    DrvScript *drv = (DrvScript *)s->mHal.drv;
    RsdCpuReference::CpuScript *cs = (RsdCpuReference::CpuScript *)drv->cs;
    cs->invokeForEach(slot, ain, aout, usr, usrLen, sc);
}

void rsdScriptInvokeForEachMulti(const Context *rsc,
                                 Script *s,
                                 uint32_t slot,
                                 const Allocation ** ains,
                                 size_t inLen,
                                 Allocation * aout,
                                 const void * usr,
                                 size_t usrLen,
                                 const RsScriptCall *sc) {

    // ARM - load CpuScript * from DrvScript structure
    DrvScript *drv = (DrvScript *)s->mHal.drv;
    RsdCpuReference::CpuScript *cs = (RsdCpuReference::CpuScript *)drv->cs;
    cs->invokeForEachMulti(slot, ains, inLen, aout, usr, usrLen, sc);
}


int rsdScriptInvokeRoot(const Context *dc, Script *s) {
    // ARM - load CpuScript * from DrvScript structure
    DrvScript *drv = (DrvScript *)s->mHal.drv;
    RsdCpuReference::CpuScript *cs = (RsdCpuReference::CpuScript *)drv->cs;
    return cs->invokeRoot();
}

void rsdScriptInvokeInit(const Context *dc, Script *s) {
    // ARM - load CpuScript * from DrvScript structure
    DrvScript *drv = (DrvScript *)s->mHal.drv;
    RsdCpuReference::CpuScript *cs = (RsdCpuReference::CpuScript *)drv->cs;
    cs->invokeInit();
}

void rsdScriptInvokeFreeChildren(const Context *dc, Script *s) {
    // ARM - load CpuScript * from DrvScript structure
    DrvScript *drv = (DrvScript *)s->mHal.drv;
    RsdCpuReference::CpuScript *cs = (RsdCpuReference::CpuScript *)drv->cs;
    cs->invokeFreeChildren();
}

void rsdScriptInvokeFunction(const Context *dc, Script *s,
                             uint32_t slot,
                             const void *params,
                             size_t paramLength) {
    // ARM - load CpuScript * from DrvScript structure
    DrvScript *drv = (DrvScript *)s->mHal.drv;
    RsdCpuReference::CpuScript *cs = (RsdCpuReference::CpuScript *)drv->cs;
    cs->invokeFunction(slot, params, paramLength);
}

void rsdScriptSetGlobalVar(const Context *dc, const Script *s,
                           uint32_t slot, void *data, size_t dataLength) {
    // ARM - load CpuScript * from DrvScript structure
    DrvScript *drv = (DrvScript *)s->mHal.drv;
    RsdCpuReference::CpuScript *cs = (RsdCpuReference::CpuScript *)drv->cs;
    cs->setGlobalVar(slot, data, dataLength);
}

void rsdScriptGetGlobalVar(const Context *dc, const Script *s,
                           uint32_t slot, void *data, size_t dataLength) {
    // ARM - load CpuScript * from DrvScript structure
    DrvScript *drv = (DrvScript *)s->mHal.drv;
    RsdCpuReference::CpuScript *cs = (RsdCpuReference::CpuScript *)drv->cs;
    cs->getGlobalVar(slot, data, dataLength);
}


void rsdScriptSetGlobalVarWithElemDims(const Context *dc, const Script *s,
                                       uint32_t slot, void *data, size_t dataLength,
                                       const android::renderscript::Element *elem,
                                       const uint32_t *dims, size_t dimLength) {
    // ARM - load CpuScript * from DrvScript structure
    DrvScript *drv = (DrvScript *)s->mHal.drv;
    RsdCpuReference::CpuScript *cs = (RsdCpuReference::CpuScript *)drv->cs;
    cs->setGlobalVarWithElemDims(slot, data, dataLength, elem, dims, dimLength);
}

void rsdScriptSetGlobalBind(const Context *dc, const Script *s, uint32_t slot, Allocation *data) {
    // ARM - load CpuScript * from DrvScript structure
    DrvScript *drv = (DrvScript *)s->mHal.drv;
    RsdCpuReference::CpuScript *cs = (RsdCpuReference::CpuScript *)drv->cs;
    cs->setGlobalBind(slot, data);
}

void rsdScriptSetGlobalObj(const Context *dc, const Script *s, uint32_t slot, ObjectBase *data) {
    // ARM - load CpuScript * from DrvScript structure
    DrvScript *drv = (DrvScript *)s->mHal.drv;
    RsdCpuReference::CpuScript *cs = (RsdCpuReference::CpuScript *)drv->cs;
    cs->setGlobalObj(slot, data);
}

void rsdScriptDestroy(const Context *dc, Script *s) {
    // ARM - load CpuScript * from DrvScript structure
    DrvScript *drv = (DrvScript *)s->mHal.drv;
    RsdCpuReference::CpuScript *cs = (RsdCpuReference::CpuScript *)drv->cs;
    delete cs;
    // ARM - delete DrvScript structure
    delete drv;
    s->mHal.drv = NULL;
}


Allocation * rsdScriptGetAllocationForPointer(const android::renderscript::Context *dc,
                                              const android::renderscript::Script *sc,
                                              const void *ptr) {
    // ARM - load CpuScript * from DrvScript structure
    DrvScript *drv = (DrvScript *)sc->mHal.drv;
    RsdCpuReference::CpuScript *cs = (RsdCpuReference::CpuScript *)drv->cs;
    return cs->getAllocationForPointer(ptr);
}

void rsdScriptUpdateCachedObject(const Context *rsc,
                                 const Script *script,
                                 rs_script *obj)
{
    obj->p = script;
#ifdef __LP64__
    obj->r = NULL;
    if (script != NULL) {
        obj->v1 = script->mHal.drv;
    } else {
        obj->v1 = NULL;
    }
    obj->v2 = NULL;
#endif
}
